package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** an efficient representation of a collection of trees (Tomita's shared packed parse forest) */
public abstract class Forest<T> /*extends PrintableTree<Forest.MyBody<T>>*/ implements Visitable<Forest.Body<T>> {

    /** assume that this forest contains exactly one tree and return it; otherwise throw an exception */
    public final Tree<T> expand1() throws Ambiguous, ParseFailed {
        Iterator<Tree<T>> it = expand(true).iterator();
        if (!it.hasNext()) throw new ParseFailed();
        return it.next();
    }

    /** expand this forest into a set of trees */
    public HashSet<Tree<T>> expand(boolean toss) {
        TreeMaker<T> ret = new TreeMaker<T>(toss);
        visit(ret, null, null);
        if (toss && ret.size() > 1) throw new Ambiguous(this);
        return ret;
    }

    static        <T> Forest<T> singleton(Input.Location loc)                       { return create(loc, null, new Forest[] { }, false, true); }
    static        <T> Forest<T> singleton(Input.Location loc, Forest<T> body)       { return create(loc, null, new Forest[] { body },  false, true); }
    static        <T> Forest<T> leaf(Input.Location loc, T tag) { return create(loc, tag, null, false, false); }
    public static <T> Forest<T> create(Input.Location loc, T tag, Forest<T>[] tokens, boolean unwrap, boolean singleton) {
        return new MyBody<T>(loc, tag, tokens, unwrap, singleton);
    }

    // Body //////////////////////////////////////////////////////////////////////////////

    protected static interface Body<T> {
        TreeMaker<T> expand(int i, TreeMaker<T> h);
    }

    protected static class MyBody<T> extends Forest<T> implements Body<T> /* extends PrintableTree<Forest<T>> implements */, Invokable<Tree<T>, TreeMaker<T>, Integer> {

        public <B,C> void visit(Invokable<Forest.Body<T>,B,C> ivbc, B b, C c) {
            ivbc.invoke(this, b, c);
        }

        private final Input.Location    location;
        private final T                 tag;
        private final Forest<T>[]       tokens;
        private final boolean           unwrap;
        private final boolean           singleton;

        private MyBody(Input.Location loc, T tag, Forest<T>[] tokens, boolean unwrap, boolean singleton) {
            this.location = loc;
            this.tag = tag;
            this.tokens = tokens==null ? emptyForestArray : new Forest[tokens.length];
            if (tokens != null) System.arraycopy(tokens, 0, this.tokens, 0, tokens.length);
            if (tokens != null) for(int i=0; i<tokens.length; i++) if (tokens[i]==null) throw new Error(i+"");
            this.unwrap = unwrap;
            this.singleton = singleton;
        }

        public TreeMaker<T> expand(int i, TreeMaker<T> h) {
            if (singleton) {
                tokens[0].visit(h, null, i);

            } else if (i==tokens.length) {
                h.add(new Tree<T>(null, tag, h.toks.toArray(tree_hint)));

            } else if (unwrap && i==tokens.length-1) {
                if (tokens[i] != null)
                    tokens[i].visit(h, null, 0);

            } else {
                //boolean hit = false;
                //tokens[i].visit(this, h, i);

                for(Tree<T> r : tokens[i].expand(h.toss)) {
                    //hit = true;
                    int old = h.toks.size();
                    h.toks.add(r);
                    expand(i+1, h);
                    while(h.toks.size() > old) h.toks.remove(h.toks.size()-1);
                }

                //if (!hit) throw new Error();
            }
            return h;
        }
        public void invoke(Tree<T> r, TreeMaker<T> h, Integer i) {
            //hit = true;
            int old = h.toks.size();
            h.toks.add(r);
            expand(i+1, h);
            while(h.toks.size() > old) h.toks.remove(h.toks.size()-1);
            //ivbc.invoke(this, b, c);
        }

        protected String  headToString()         { return null; }
        protected String  headToJava()           { return null; }
        protected String  left()                 { return "{"; }
        protected String  right()                { return "}"; }
        protected boolean ignoreSingleton()      { return false; }
    }


    // Ref //////////////////////////////////////////////////////////////////////////////

    /**
     *  This class represents a partially complete collection of
     *  forests to be viewed as a forest at some later date; once
     *  viewed, it becomes immutable
     */
    static class Ref<T> extends Forest<T> {
        private FastSet<Forest<T>> hp = new FastSet<Forest<T>>();
        public Ref() { }
        public void merge(Forest p) { if (p!=this) hp.add(p, true); }
        public <B,C> void visit(Invokable<Forest.Body<T>,B,C> ivbc, B b, C c) {
            if (hp==null) return;
            for(Forest<T> f : hp)
                f.visit(ivbc, b, c);
        }
        public Forest resolve() { return this; }
    }

    public abstract <B,C> void visit(Invokable<Forest.Body<T>,B,C> ivbc, B b, C c);
    private static class TreeMaker<T> extends HashSet<Tree<T>> implements Invokable<Forest.Body<T>,Boolean,Integer> {
        public ArrayList<Tree<T>> toks = new ArrayList<Tree<T>>();
        private boolean toss;
        public TreeMaker(boolean toss) { this.toss = toss; }
        public void invoke(Forest.Body<T> bod, Boolean o, Integer i) {
            if (i==null) {
                ArrayList<Tree<T>> toks = this.toks;
                this.toks = new ArrayList<Tree<T>>();
                bod.expand(0, this);
                this.toks = toks;
            } else {
                bod.expand(i, this);
            }
        }
    }

    // Statics //////////////////////////////////////////////////////////////////////////////

    private static Tree[] tree_hint = new Tree[0];
    private static final Forest[] emptyForestArray = new Forest[0];

    protected String  headToString()    { return null; }
    protected String  headToJava()      { return null; }
    protected String  left()            { return "<?"; }
    protected String  right()           { return "?>"; }
    protected boolean ignoreSingleton() { return true; }
}
